<?php

namespace UNSProjectApp;

use BrowserDetector\Browser;
use BrowserDetector\Device;
use BrowserDetector\Os;
use UNSProjectApp\Libraries\JWT;

class UnsApp
{
    const PRIVATE_KEY_BITS = 4096;
    const API_URL = 'https://uns-poc-comptroller.web.app/api';
    const GUARDIAN_API_URL = 'https://guardian.poc.unsproject.com';
    const DEFAULT_ATTESTATION_TYPE = 'email-verified';

    /**
     * @param string $serviceTicket
     * @return string
     */
    public static function generateQRCodeLink($serviceTicket)
    {
        return self::GUARDIAN_API_URL .'/processAuthenticationTicket?serviceTicket=' . $serviceTicket;
    }

    public static function checkPluginRequirements(){
        if (!extension_loaded('openssl')) {
            die('Openssl extension is missing.');
        }

        if(!extension_loaded('curl')){
            die('Curl extension is missing.');
        }

        if (substr(site_url(), 0, strlen('https')) !== 'https') {
            die('This plugin works only on https://.');
        }
    }

    /**
     * @param SiteOptions $siteOption
     *
     * @return boolean
     * @throws Libraries\JWTException|UnsAppException
     */
    public static function deleteService($siteOption)
    {
        if(empty($siteOption->getValue('private_key'))){
            throw new UnsAppException('Missing plugin config. Please contact UNSproject support team to help you.');
        }

        $parameters = [
            'body'  => [
                    'serviceNamespace' => $siteOption->getValue('uniqueId')
            ],
            'headers' => [
                'authorization' =>'Basic' . ' ' . JWT::encode(
                    [
                        'delete' => $siteOption->getValue('uniqueId')
                    ],
                    base64_decode($siteOption->getValue('private_key')),
                    'RS256'
                )
            ]
        ];

        $url = self::API_URL. '/services';
        ServerCall::delete($url, $parameters, $statusCode, $result);

         if(!in_array($statusCode, [200, 201])){
             throw new UnsAppException(
                 sprintf(
                     'Unable to delete the service. Status code: %s. Message: %s',
                     $statusCode,
                     strip_tags($result)
                 )
             );
         }
    }
    
    /**
     * @param SiteOptions $siteOption
     * @return SiteOptions
     * @throws \Exception
     */
    public static function initializeRegisterProcess($siteOption)
    {
        list($privateKey, $publicKey) = self::generateNewPrivatePublicKeys();

        $siteOption->setValue('public_key', base64_encode($publicKey));
        $siteOption->setValue('private_key', base64_encode($privateKey));
        $siteOption->setValue('site_url', site_url());
        $siteOption->setValue('uniqueId', str_replace(['https://','http://','/'],'', site_url()));

        //Register website
        $params = [
            'serviceNamespace' => $siteOption->getValue('uniqueId'),
            'domain' => $siteOption->getValue('site_url'),
            'contactName' => $siteOption->getValue('contactName'),
            'email'       => $siteOption->getValue('email'),
            'phoneNumber' => $siteOption->getValue('phoneNumber'),
            'callbackPath' => UnsWordPressAPI::getCallBackPath(),
            'domainValidationPath' => UnsWordPressAPI::getValidationPath(),
            'publicKey' => base64_decode($siteOption->getValue('public_key'))
        ];

        $url = self::API_URL . '/services/register';
        $response = ServerCall::post($url, ['body' => $params] , $statusCode, $plainResult);

        if (!isset($response['validationCode'])) {
            throw new \Exception('Wrong website configuration. Unable to get the validation code.' . $plainResult);
        }
        $siteOption->setValue('validationCode', $response['validationCode']);
        if(isset($response['gatekeeperPublicKey'])){
            $siteOption->setValue('gatekeeperPublicKey', $response['gatekeeperPublicKey']);
        }
        $siteOption->setValue('registeredSite', true);
        $siteOption->save();

        return $siteOption;
    }

    /**
     * @param SiteOptions $siteOption
     * @throws \Exception
     */
    public static function initializeSiteValidation($siteOption){
        $expectedResult = 'OK';

        $params = [
            'serviceNamespace' => $siteOption->getValue('uniqueId'),
            'validationCode' => $siteOption->getValue('validationCode'),
        ];

        $result  = ServerCall::post(self::API_URL . '/services/validate', ['body' => $params] , $statusCode, $plainResult);
        if(empty($result) && !empty($plainResult) && $plainResult !== $expectedResult){
            throw  new UnsAppException('Error while validating the website. '.ucfirst(strip_tags($plainResult)), UnsAppException::VALIDATION_ERROR);
        }
    }

    /**
     * @param SiteOptions $pluginOption
     * @param bool $usedInFrontend
     * @return string|mixed
     * @throws \Exception
     */
    public static function requestTicket($pluginOption, $usedInFrontend = false)
    {
        $defaultAttestationType = $pluginOption->getValue('default_attestation_type') !== null
            ? $pluginOption->getValue('default_attestation_type')
            : self::DEFAULT_ATTESTATION_TYPE;

        //https://github.com/sinergi/php-browser-detector
        $browser = new Browser($_SERVER['HTTP_USER_AGENT']);
        $device = new Device($_SERVER['HTTP_USER_AGENT']);
        $os = new Os($_SERVER['HTTP_USER_AGENT']);

        //Ticket request
        $params = [
            'serviceNamespace' => $pluginOption->getValue('uniqueId'),
            'attestationTypeRequired' =>  $defaultAttestationType,
            "deviceInformation" =>  [
                "deviceType" =>  $os->isMobile() ? 'mobile' : 'browser',
                "deviceBrand" =>  'Unknown',
                "deviceName" => $device->getName(),
                "browserName" => $browser->getName(),
                "browserVersion" => $browser->getVersion(),
                "osName" =>  $os->getName(),
                "osVersion" => $os->getVersion(),
                "fingerPrint" => 0
            ]
        ];

        $url = self::API_URL . '/tickets/request';
        $response = ServerCall::post($url, ['body' => $params], $statusCode, $plainTextResult);

        if (!isset($response['authenticationTicket'])) {
            throw new \Exception('Unable to generate ticket.'. strip_tags($plainTextResult));
        }

        try {
            $authenticationTicketParts = explode('.', $response['authenticationTicket']);
            if(!isset($authenticationTicketParts[1])){
                throw new \Exception('Invalid JWT parts.');
            }

            $payload = json_decode(base64_decode($authenticationTicketParts[1]), true);

            if(!isset($payload['ticketID'])){
                throw  new \Exception('Missing ticketID from payload.');
            }

            $ticketId = $payload['ticketID'];
            $jwtPayload = [
                'serviceValidatedAt' => date(\DateTime::ISO8601, time()),
                'authenticationTicket' => $response['authenticationTicket']
            ];

            $signetJWT = JWT::encode($jwtPayload,
                base64_decode($pluginOption->getValue('private_key')),
                'RS256'
            );
        } catch (\Exception $e) {
            throw  new \Exception($e);
        }

        $databaseService = new DatabaseService();
        if($usedInFrontend){
            $databaseService->saveIntoTickets($ticketId,0);
        }else {
            $usersID = $databaseService->saveIntoUsersTable(get_current_user_id(),$defaultAttestationType, null);
            $databaseService->saveIntoTickets($ticketId, $usersID);
        }

        return  $signetJWT;
    }

    /**
     * @return array [$privateKey, $publicKey]
     */
    private static function generateNewPrivatePublicKeys()
    {
        $keyGenerator = openssl_pkey_new(array(
            "digest_alg" => "sha512",
            'private_key_bits' => self::PRIVATE_KEY_BITS,
            'private_key_type' => OPENSSL_KEYTYPE_RSA
        ));

        openssl_pkey_export($keyGenerator, $privateKey);
        $details = openssl_pkey_get_details($keyGenerator);
        $publicKey = $details['key'];
        openssl_pkey_free($keyGenerator);

        return [$privateKey, $publicKey];
    }

    /**
     * @return string[]
     */
    public static function getAttestationTypes(){
        return [
            'none' => 'None',
            'email-not-verified' => 'Email not Verified',
            'email-verified' => 'Email verified',
            'pinned-with-pki' => 'Pinned with PKI',
            'pki' => 'PKI'
        ];
    }

    /**
     * @param string $attestationKeyToSearch
     * @return array
     */
    public static function getLowerAttestations($attestationKeyToSearch)
    {
        $allAttestations = array_keys(self::getAttestationTypes());

        $index = array_search($attestationKeyToSearch, $allAttestations);
        if($index === false){
            return [];
        }

        return array_splice($allAttestations,$index);
    }

}
